#!/usr/bin/env python3

import time
from typing import Optional

from pycozo.client import Client

from .schema import MANAGER_TABLE_NAME
from .types import MigrationModule
from .utils.console import fail, info, warn
from .utils.fn import paired


def add_migration(
    client: Client, id: Optional[str], previous_id: Optional[str], created_at: float
) -> None:
    """Add a migration to the migrations manager table."""

    time.sleep(0.01)  # Ensure that the migrated_at is unique
    client.insert(
        MANAGER_TABLE_NAME,
        {
            "created_at": created_at,
            "previous_id": previous_id,
            "id": id,
        },
    )


def apply(
    client: Client,
    migration1: MigrationModule,
    migration2: MigrationModule,
    down: bool = False,
) -> str:
    """Apply a migration and return the operation name."""
    # Up: migration1 -> migration2 (run migration2.up)
    # Down: migration1 -> migration2 (run migration1.down)
    op = migration1.down if down else migration2.up

    op_name = (
        f"{migration1.MIGRATION_ID}::down" if down else f"{migration2.MIGRATION_ID}::up"
    )

    previous_id = migration1.MIGRATION_ID
    id = migration2.MIGRATION_ID
    created_at = migration2.CREATED_AT
    assert id or previous_id, "Can't migrate from None to None"

    op(client)

    add_migration(
        client,
        created_at=created_at,
        id=id,
        previous_id=previous_id,
    )

    return op_name


def apply_migrations(
    client: Client,
    migrations: list[MigrationModule],
    down: bool = False,
    _rolling_back: bool = False,
    verbose: bool = False,
    from_cli: bool = True,
) -> tuple[bool, Optional[Exception]]:
    currently_applied = []
    need_to_roll_back = False
    migration_error = None

    for from_, to in paired(migrations):
        try:
            op_name = apply(client, from_, to, down=down)
            from_cli and verbose and info(f"Executed `{op_name}`.")
            currently_applied.append(from_)
        except Exception as e:
            if _rolling_back:
                # We're already rolling back, so we can't roll back again
                # Just fail
                from_cli and fail(f"Rollback failed: {e}")
                raise e

            migration_error = e
            need_to_roll_back = True
            break

    if need_to_roll_back:
        from_cli and migration_error and warn(f"Error: {migration_error}")

        # Roll back the migrations that were applied
        # in reverse order
        from_cli and warn("Migration failed. Reverting...")

        apply_migrations(
            client,
            currently_applied[::-1],
            down=(not down),
            _rolling_back=True,
            verbose=True,
            from_cli=from_cli,
        )

        return (False, migration_error)

    return (True, None)



from pycozo.client import Client

from ..schema import create_schema, schema_exists


def init(client: Client):
    """Initialize the database with the migration manager table."""

    assert not schema_exists(client), "Schema already exists."
    create_schema(client)



from pycozo.client import Client

from ..queries import get_history


def history(client: Client):
    """Get the migration history in the database as a pandas dataframe."""

    history = get_history(client)
    return history



from typing import Optional

from rich import print
import typer

from ..queries import get_latest_migration
from ..types import Migration
from ..utils.console import fail

from .main import app


@app.command()
def status(ctx: typer.Context):
    """Display the current migration status."""

    # Validate client
    ctx.obj.check_client()

    client = ctx.obj.client

    last_migration: Optional[Migration] = get_latest_migration(client)
    if last_migration is None:
        fail("No migrations have been applied yet.")

    else:
        print(last_migration.to_table())



from pathlib import Path
from typing import Optional
from typing_extensions import Annotated

import typer

from ..context import AppContext
from ..types import CozoConnectionOptions, EngineType

app = typer.Typer(pretty_exceptions_show_locals=False)


@app.callback()
def main(
    ctx: typer.Context,
    migrations_dir: Annotated[
        Path,
        typer.Option(
            "--migrations-dir",
            "-d",
            help="Directory to use for looking up migration files.",
            envvar="COZO_MIGRATIONS_DIR",
        ),
    ] = Path("./migrations"),
    engine: Annotated[
        EngineType,
        typer.Option("--engine", "-e", help="Engine to use", envvar="COZO_ENGINE"),
    ] = EngineType.sqlite,
    path: Annotated[
        Optional[Path],
        typer.Option(
            "--path",
            "-f",
            help="Database file (not applicable for mem or http engines)",
            envvar="COZO_PATH",
        ),
    ] = None,
    host: Annotated[
        str,
        typer.Option(
            "--host",
            "-h",
            help="Host to connect to (http engine only)",
            envvar="COZO_HOST",
        ),
    ] = "http://127.0.0.1:9070",
    auth: Annotated[
        Optional[str],
        typer.Option(
            help="Auth header (http engine only)", envvar="COZO_AUTH", show_envvar=False
        ),
    ] = None,
    verbose: Annotated[bool, typer.Option("--verbose", "-v")] = False,
):
    """
    A simple migration tool for Cozo databases.
    """

    # Prepare context
    options: CozoConnectionOptions = {"host": host, "auth": auth}

    app_context: AppContext = AppContext(
        engine=engine,
        path=path,
        options=options,
        migrations_dir=migrations_dir,
        verbose=verbose,
    )

    ctx.obj = app_context



from datetime import datetime, timezone


# Note: datetime.utcnow() returns a datetime object with tzinfo set to None
# Which is not the same as datetime.now(timezone.utc)
# See: https://medium.com/@life-is-short-so-enjoy-it/python-datetime-utcnow-maybe-no-more-for-me-221795e8ddbf
def utcnow():
    return datetime.now(timezone.utc)



#!/usr/bin/env python3

from pycozo.client import Client

from .utils.fn import doesnt_throw

MANAGER_TABLE_NAME: str = "cozo_migrate_manager"

check_schema_query = f"""
?[id] := *{MANAGER_TABLE_NAME} {{ id }}
:limit 1
"""

create_schema_query = f"""
:create {MANAGER_TABLE_NAME} {{
    migrated_at_ms: Validity default [floor(now() * 1000), true] =>
    id: String?,
    previous_id: String?,
    created_at: Float?,
}}"""


@doesnt_throw
def schema_exists(client: Client) -> bool:
    # Ensure the migrate schema exists
    client.run(check_schema_query)
    return True


def create_schema(client: Client) -> None:
    client.run(create_schema_query)



from pathlib import Path
import re

from ..files import get_migration_files
from ..template import migration_template
from ..utils.datetime import utcnow


def create(
    id: str,
    migration_dir: str | Path,
) -> None:
    """
    Create a new migration file in the migrations directory.
    Format: {migration_dir}/migrate_<timestamp>_<id>.py
    """

    # Create migrations_dir if it doesn't exist
    migrations_dir = Path(migration_dir)
    migrations_dir.mkdir(parents=True, exist_ok=True)

    # id can only contain letters, numbers, and underscores
    # can only start with a lowercase letter
    regex = r"^[a-z][a-zA-Z0-9_]*$"
    assert re.match(
        regex, id
    ), "`id` can only contain letters, numbers, and underscores"

    # Check that a migration with same id doesn't already exist
    migration_files = get_migration_files(migrations_dir)
    any_existing = any(m.id == id for m in migration_files)

    assert not any_existing, f"A migration with id `{id}` already exists."

    # Create the migration file
    ts = utcnow().timestamp()
    path = migrations_dir / f"migrate_{int(ts)}_{id}.py"

    with open(path, "w") as f:
        f.write(migration_template.format(migration_id=id, created_at=ts))



import re
from typing_extensions import Annotated

import typer

from ..files import get_migration_files
from ..template import migration_template
from ..utils.console import fail, success
from ..utils.datetime import utcnow

from .main import app


@app.command()
def create(
    ctx: typer.Context,
    id: Annotated[
        str,
        typer.Argument(
            ...,
            help="A short, descriptive, alphanumeric id for the migration. Only letters, numbers, and underscores are allowed.",
        ),
    ],
    confirm: Annotated[bool, typer.Option("--yes", "-y")] = False,
):
    """
    Create a new migration file in the migrations directory.
    Format: {migration_dir}/migrate_<timestamp>_<id>.py
    """

    migrations_dir = ctx.obj.migrations_dir

    # Create migrations_dir if it doesn't exist
    migrations_dir.mkdir(parents=True, exist_ok=True)

    # id can only contain letters, numbers, and underscores
    # can only start with a lowercase letter
    regex = r"^[a-z][a-zA-Z0-9_]*$"
    if not re.match(regex, id):
        fail("`id` can only contain letters, numbers, and underscores")

    # Check that a migration with same id doesn't already exist
    migration_files = get_migration_files(migrations_dir)
    any_existing = any(m.id == id for m in migration_files)
    if any_existing:
        fail(f"A migration with id `{id}` already exists.")

    # Create the migration file
    ts = utcnow().timestamp()
    path = migrations_dir / f"migrate_{int(ts)}_{id}.py"

    # Confirm
    if not confirm:
        confirmed = typer.confirm(f"Writing '{path}' Confirm?")
        if not confirmed:
            raise typer.Abort()

    with open(path, "w") as f:
        f.write(migration_template.format(migration_id=id, created_at=ts))

    success(f"Created migration at: {path}")



#!/usr/bin/env python3

from dataclasses import dataclass
from pathlib import Path
from typing import Optional

from pycozo.client import Client

from .schema import schema_exists
from .types import EngineType, CozoConnectionOptions
from .utils.client import is_connected
from .utils.console import fail


@dataclass
class AppContext:
    engine: EngineType
    path: Path
    options: CozoConnectionOptions
    migrations_dir: Path
    verbose: bool
    _client: Optional[Client] = None

    @property
    def client(self) -> Client:
        if self._client is None:
            self._client = Client(self.engine, str(self.path), options=self.options)

        return self._client

    def ensure_connected(self) -> None:
        if not is_connected(self.client):
            fail("Could not connect to database")

    def ensure_schema(self) -> None:
        # Ensure schema exists
        if not schema_exists(self.client):
            fail("Schema does not exist. Run `cozo-migrate init` first.")

    def validate_options(self) -> None:
        # Validate options
        if self.engine in (EngineType.sqlite, EngineType.rocksdb):
            if not self.path:
                fail(
                    "`--path / -f` is required for sqlite and rocksdb engines"
                    "\n\n"
                    "Example:\n"
                    "$ cozo-migrate -e sqlite -f ./mydb.db [COMMAND] [OPTIONS]"
                )

            self.path = self.path.resolve()

        if self.engine == EngineType.http:
            if not (self.options and self.options.get("host")):
                fail("`--host / -h` is required for http engine")

    def check_client(self) -> None:
        self.validate_options()
        self.ensure_connected()
        self.ensure_schema()



from importlib import metadata

from ..utils.console import info
from .main import app


@app.command()
def version():
    """Display the current version."""

    info(f"cozo-migrate {metadata.version('cozo-migrate')}")



from pathlib import Path
from typing import Optional

from pycozo.client import Client

from ..files import get_migration_files
from ..modules import get_adjacent_migrations
from ..ops import apply_migrations
from ..queries import get_latest_migration
from ..types import Migration, MigrationModule, MigrationModuleInfo


def apply(
    client: Client,
    migrations_dir: str | Path,
    steps: int = 1,
    all_: bool = False,
    down: bool = False,
):
    """
    Apply migrations to the database.
    You can specify the number of steps to apply and the direction.
    """

    # Get the current migration
    current_migration: Optional[Migration] = get_latest_migration(client)
    current_id: Optional[str] = current_migration.id if current_migration else None

    if down and current_id is None:
        raise ValueError("No migrations to rollback.")

    # Find migrations in migrations_dir
    migration_files: list[MigrationModuleInfo] = get_migration_files(migrations_dir)
    num_migrations_available = len(migration_files)
    assert num_migrations_available > 0, "No migrations found."

    # If all migrations are requested, set steps to the number of migrations available
    if all_:
        steps = num_migrations_available

    # Get the next steps migrations from starting_id (can be negative)
    migrations: list[MigrationModule] = get_adjacent_migrations(
        migration_files, current_id, steps, down
    )

    assert len(migrations) > 0, "No migrations to apply."

    applied_successfully, migration_error = apply_migrations(
        client, migrations, down=down, verbose=False, from_cli=False
    )

    if not applied_successfully:
        raise ValueError("Migration failed.") from migration_error



from typing import Optional

from pycozo.client import Client

from ..queries import get_latest_migration
from ..types import Migration


def status(client: Client):
    """Display the current migration status."""

    last_migration: Optional[Migration] = get_latest_migration(client)
    assert last_migration is not None, "No migrations have been applied yet."

    return last_migration



from typing import Optional

from rich import print
import typer

from ..queries import get_current_schema
from ..types import Migration
from ..utils.console import fail
from ..utils.rich import df_to_table, pretty_transforms

from .main import app


@app.command()
def show(ctx: typer.Context):
    """Show the current schema in the database."""

    # Validate client
    ctx.obj.check_client()

    client = ctx.obj.client

    db_schema = get_current_schema(client)

    for relation, columns in db_schema.items():
        print(f"[bold][blue]\[{relation}][/blue][/bold]")
        print(df_to_table(columns))
        print()



from typing import Optional
from typing_extensions import Annotated

import typer

from ..files import get_migration_files
from ..modules import get_adjacent_migrations
from ..ops import apply_migrations
from ..queries import get_latest_migration
from ..types import Migration, MigrationModule, MigrationModuleInfo
from ..utils.console import fail, success, info

from .main import app


@app.command()
def apply(
    ctx: typer.Context,
    steps: Annotated[int, typer.Argument()] = 1,
    down: Annotated[bool, typer.Option("--down")] = False,
    all_: Annotated[bool, typer.Option("--all", "-a")] = False,
    confirm: Annotated[bool, typer.Option("--yes", "-y")] = False,
    compact: Annotated[bool, typer.Option("--compact")] = False,
):
    """
    Apply migrations to the database.
    You can specify the number of steps to apply and the direction.
    """

    # Validate client
    ctx.obj.check_client()

    client = ctx.obj.client
    migrations_dir = ctx.obj.migrations_dir
    verbose = ctx.obj.verbose

    # Get the current migration
    current_migration: Optional[Migration] = get_latest_migration(client)
    current_id: Optional[str] = current_migration.id if current_migration else None

    if down and current_id is None:
        fail("No migrations to rollback.")

    # Find migrations in migrations_dir
    migration_files: list[MigrationModuleInfo] = get_migration_files(migrations_dir)
    num_migrations_available = len(migration_files)
    if num_migrations_available == 0:
        fail("No migrations found.")

    # If all migrations are requested, set steps to the number of migrations available
    if all_:
        steps = num_migrations_available

    # Get the next steps migrations from starting_id (can be negative)
    migrations: list[MigrationModule] = get_adjacent_migrations(
        migration_files, current_id, steps, down
    )

    if len(migrations) == 0:
        fail("No migrations to apply.")

    # Print the migration path
    joiner = "[yellow] → [/yellow]"
    migration_ids = [m.MIGRATION_ID or "[NONE]" for m in migrations]

    info(f"Migrate path: {joiner.join(migration_ids)}")

    # Confirm
    if not confirm:
        confirmed = typer.confirm("Are you sure you want to apply these migrations?")
        if not confirmed:
            raise typer.Abort()

    info("Migrating the database...")
    applied_successfully, migration_error = apply_migrations(
        client, migrations, down=down, verbose=verbose, from_cli=True
    )

    if compact:
        client.run("::compact")
        info("Database compaction initiated.")

    if applied_successfully:
        success("Database migrated.")
    else:
        fail("Migration failed.", error=migration_error)



#!/usr/bin/env python3

from typing import Optional

import pandas as pd
from pycozo.client import Client

from .schema import MANAGER_TABLE_NAME
from .types import Migration

get_first_query = f"""
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *{MANAGER_TABLE_NAME} {{
        migrated_at_ms: validity,
        id,
        created_at,
        previous_id,
    }},
    migrated_at = to_int(validity) / 1000,
    previous_id = null
"""

get_latest_query = f"""
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *{MANAGER_TABLE_NAME} {{
        migrated_at_ms: validity,
        id,
        created_at,
        previous_id,
        @ "NOW"
    }}, migrated_at = to_int(validity) / 1000
"""

get_history_query = f"""
?[
    migrated_at,
    id,
    previous_id,
    created_at,
] := *{MANAGER_TABLE_NAME} {{
        migrated_at_ms: validity,
        id,
        created_at,
        previous_id,
    }}, migrated_at = to_int(validity) / 1000

:sort migrated_at
"""

get_migration_path_query = f"""
edges[
    previous_id,
    id,
] := *{MANAGER_TABLE_NAME} {{
        id,
        previous_id,
    }}

starting[id] <- [[null]]
latest[id] := *{MANAGER_TABLE_NAME} {{ id @ "NOW" }}

?[start, goal, path] <~ ShortestPathBFS(
    edges[from, to],
    starting[start_idx],
    latest[goal_idx],
)
"""


def get_history(client: Client) -> pd.DataFrame:
    results = client.run(get_history_query)
    return results


def get_migration_path(client: Client) -> list[str]:
    results = client.run(get_migration_path_query)
    assert len(results) <= 1, "More than one path returned"

    if len(results) == 0:
        return []

    path = results["path"][0]
    return path


def get_latest_migration(client: Client) -> Optional[Migration]:
    migration = Migration.get_first(client.run(get_latest_query))
    return migration


def get_first_migration(client: Client) -> Optional[Migration]:
    migration = Migration.get_first(client.run(get_first_query))
    return migration


def get_current_schema(client: Client) -> pd.DataFrame:
    all_relations = client.run("::relations")
    normal_relations = all_relations[all_relations["access_level"] != "index"][
        ["name", "description"]
    ]

    relation_wise_columns = {
        relation: client.run(f"::columns {relation}").drop("index", axis=1)
        for relation in normal_relations["name"]
        if relation != MANAGER_TABLE_NAME
    }

    return relation_wise_columns



from rich import print
import typer

from ..queries import get_history
from ..utils.rich import df_to_table, pretty_transforms

from .main import app


@app.command()
def history(ctx: typer.Context):
    """Display the migration history in the database as a pretty table."""

    # Validate client
    ctx.obj.check_client()

    client = ctx.obj.client

    history = get_history(client).transform(pretty_transforms)
    history_table = df_to_table(history)

    print(history_table)
